#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This is the unittest of SRF class.
"""
from __future__ import division, absolute_import, print_function

import unittest
import numpy as np
from gstools import SRF, Gaussian
from gstools import transform as tf


class TestSRF(unittest.TestCase):
    def setUp(self):
        self.cov_model = Gaussian(dim=2, var=1.5, len_scale=4.0, mode_no=100)
        self.mean = 0.3
        self.mode_no = 100

        self.seed = 825718662
        self.x_grid = np.linspace(0.0, 12.0, 48)
        self.y_grid = np.linspace(0.0, 10.0, 46)
        self.z_grid = np.linspace(0.0, 10.0, 40)

        self.x_grid_c = np.linspace(-6.0, 6.0, 8)
        self.y_grid_c = np.linspace(-6.0, 6.0, 8)
        self.z_grid_c = np.linspace(-6.0, 6.0, 8)

        rng = np.random.RandomState(123018)
        self.x_tuple = rng.uniform(0.0, 10, 100)
        self.y_tuple = rng.uniform(0.0, 10, 100)
        self.z_tuple = rng.uniform(0.0, 10, 100)

    def test_shape_1d(self):
        self.cov_model.dim = 1
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_str = srf([self.x_grid], seed=self.seed, mesh_type="structured")
        field_unstr = srf(
            [self.x_tuple], seed=self.seed, mesh_type="unstructured"
        )
        self.assertEqual(field_str.shape, (len(self.x_grid),))
        self.assertEqual(field_unstr.shape, (len(self.x_tuple),))

    def test_shape_2d(self):
        self.cov_model.dim = 2
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_str = srf(
            (self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured"
        )
        field_unstr = srf(
            (self.x_tuple, self.y_tuple),
            seed=self.seed,
            mesh_type="unstructured",
        )
        self.assertEqual(field_str.shape, (len(self.x_grid), len(self.y_grid)))
        self.assertEqual(field_unstr.shape, (len(self.x_tuple),))

    def test_shape_3d(self):
        self.cov_model.dim = 3
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_str = srf(
            (self.x_grid, self.y_grid, self.z_grid),
            seed=self.seed,
            mesh_type="structured",
        )
        field_unstr = srf(
            (self.x_tuple, self.y_tuple, self.z_tuple),
            seed=987654,
            mesh_type="unstructured",
        )
        self.assertEqual(
            field_str.shape,
            (len(self.x_grid), len(self.y_grid), len(self.z_grid)),
        )
        self.assertEqual(field_unstr.shape, (len(self.x_tuple),))

    def test_anisotropy_2d(self):
        self.cov_model.dim = 2
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_iso = srf(
            (self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured"
        )
        self.cov_model.anis = 0.5
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_aniso = srf(
            (self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured"
        )
        self.assertAlmostEqual(field_iso[0, 0], field_aniso[0, 0])
        self.assertAlmostEqual(field_iso[0, 4], field_aniso[0, 2])
        self.assertAlmostEqual(field_iso[0, 10], field_aniso[0, 5])

    def test_anisotropy_3d(self):
        self.cov_model.dim = 3
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_iso = srf(
            (self.x_grid, self.y_grid, self.z_grid),
            seed=self.seed,
            mesh_type="structured",
        )
        self.cov_model.anis = (0.5, 4.0)
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_aniso = srf(
            (self.x_grid, self.y_grid, self.z_grid),
            seed=self.seed,
            mesh_type="structured",
        )
        self.assertAlmostEqual(field_iso[0, 0, 0], field_aniso[0, 0, 0])
        self.assertAlmostEqual(field_iso[0, 4, 0], field_aniso[0, 2, 0])
        self.assertAlmostEqual(field_iso[0, 10, 0], field_aniso[0, 5, 0])
        self.assertAlmostEqual(field_iso[0, 0, 0], field_aniso[0, 0, 0])
        self.assertAlmostEqual(field_iso[0, 0, 1], field_aniso[0, 0, 4])
        self.assertAlmostEqual(field_iso[0, 0, 3], field_aniso[0, 0, 12])

    def test_rotation_unstruct_2d(self):
        self.cov_model.dim = 2
        x_len = len(self.x_grid_c)
        y_len = len(self.y_grid_c)
        x_u, y_u = np.meshgrid(self.x_grid_c, self.y_grid_c)
        x_u = np.reshape(x_u, x_len * y_len)
        y_u = np.reshape(y_u, x_len * y_len)

        self.cov_model.anis = 0.25
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)

        field = srf((x_u, y_u), seed=self.seed, mesh_type="unstructured")
        field_str = np.reshape(field, (y_len, x_len))

        self.cov_model.angles = -np.pi / 2.0
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_rot = srf((x_u, y_u), seed=self.seed, mesh_type="unstructured")
        field_rot_str = np.reshape(field_rot, (y_len, x_len))

        self.assertAlmostEqual(field_str[0, 0], field_rot_str[-1, 0])
        self.assertAlmostEqual(field_str[1, 2], field_rot_str[-3, 1])

    def test_rotation_struct_2d(self):
        self.cov_model.dim = 2
        self.cov_model.anis = 0.25
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field = srf(
            (self.x_grid_c, self.y_grid_c),
            seed=self.seed,
            mesh_type="structured",
        )

        self.cov_model.angles = -np.pi / 2.0
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_rot = srf(
            (self.x_grid_c, self.y_grid_c),
            seed=self.seed,
            mesh_type="structured",
        )

        self.assertAlmostEqual(field[0, 0], field_rot[0, -1])
        self.assertAlmostEqual(field[1, 2], field_rot[2, 6])

    def test_rotation_unstruct_3d(self):
        self.cov_model = Gaussian(
            dim=3, var=1.5, len_scale=4.0, anis=(0.25, 0.5), mode_no=100
        )
        x_len = len(self.x_grid_c)
        y_len = len(self.y_grid_c)
        z_len = len(self.z_grid_c)
        x_u, y_u, z_u = np.meshgrid(
            self.x_grid_c, self.y_grid_c, self.z_grid_c
        )
        x_u = np.reshape(x_u, x_len * y_len * z_len)
        y_u = np.reshape(y_u, x_len * y_len * z_len)
        z_u = np.reshape(z_u, x_len * y_len * z_len)

        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field = srf((x_u, y_u, z_u), seed=self.seed, mesh_type="unstructured")
        field_str = np.reshape(field, (y_len, x_len, z_len))

        self.cov_model.angles = (-np.pi / 2.0, -np.pi / 2.0)
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_rot = srf(
            (x_u, y_u, z_u), seed=self.seed, mesh_type="unstructured"
        )
        field_rot_str = np.reshape(field_rot, (y_len, x_len, z_len))

        self.assertAlmostEqual(field_str[0, 0, 0], field_rot_str[-1, -1, 0])
        self.assertAlmostEqual(field_str[1, 2, 0], field_rot_str[-3, -1, 1])
        self.assertAlmostEqual(field_str[0, 0, 1], field_rot_str[-1, -2, 0])

    def test_rotation_struct_3d(self):
        self.cov_model.dim = 3
        self.cov_model.anis = 0.25
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field = srf(
            (self.x_grid_c, self.y_grid_c, self.z_grid_c),
            seed=self.seed,
            mesh_type="structured",
        )

        self.cov_model.angles = -np.pi / 2.0
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_rot = srf(
            (self.x_grid_c, self.y_grid_c, self.z_grid_c),
            seed=self.seed,
            mesh_type="structured",
        )

        self.assertAlmostEqual(field[0, 0, 0], field_rot[0, 7, 0])
        self.assertAlmostEqual(field[0, 0, 1], field_rot[0, 7, 1])

        self.cov_model.angles = (0, -np.pi / 2.0)
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field_rot = srf(
            (self.x_grid_c, self.y_grid_c, self.z_grid_c),
            seed=self.seed,
            mesh_type="structured",
        )

        self.assertAlmostEqual(field[0, 0, 0], field_rot[7, 0, 0])
        self.assertAlmostEqual(field[0, 1, 0], field_rot[7, 1, 0])
        self.assertAlmostEqual(field[1, 1, 0], field_rot[7, 1, 1])

    def test_calls(self):
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        field = srf((self.x_tuple, self.y_tuple), seed=self.seed)
        field2 = srf.unstructured((self.x_tuple, self.y_tuple), seed=self.seed)
        self.assertAlmostEqual(field[0], srf.field[0])
        self.assertAlmostEqual(field[0], field2[0])
        field = srf(
            (self.x_tuple, self.y_tuple),
            seed=self.seed,
            mesh_type="structured",
        )
        field2 = srf.structured((self.x_tuple, self.y_tuple), seed=self.seed)
        self.assertAlmostEqual(field[0, 0], srf.field[0, 0])
        self.assertAlmostEqual(field[0, 0], field2[0, 0])

    def test_transform(self):
        self.cov_model.dim = 2
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        srf((self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured")
        tf.normal_force_moments(srf)  # force ergodicity of the given field
        self.assertAlmostEqual(srf.field.mean(), srf.mean)
        self.assertAlmostEqual(srf.field.var(), srf.model.var)
        tf.zinnharvey(srf)  # make high values mostly connected
        tf.normal_force_moments(srf)  # force ergodicity of the given field
        tf.normal_to_lognormal(srf)  # log-normal
        srf((self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured")
        tf.normal_to_arcsin(srf)
        srf((self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured")
        tf.normal_to_uquad(srf)
        srf((self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured")
        tf.normal_to_uniform(srf)
        srf((self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured")
        tf.binary(srf)
        srf((self.x_grid, self.y_grid), seed=self.seed, mesh_type="structured")
        tf.boxcox(srf)

    def test_incomprrandmeth(self):
        self.cov_model = Gaussian(dim=2, var=0.5, len_scale=1.0, mode_no=100)
        srf = SRF(
            self.cov_model,
            mean=self.mean,
            mode_no=self.mode_no,
            generator="IncomprRandMeth",
            mean_velocity=0.5,
        )
        field = srf((self.x_tuple, self.y_tuple), seed=476356)
        self.assertAlmostEqual(field[0, 0], 1.23693272)
        self.assertAlmostEqual(field[0, 1], 0.89242284)
        field = srf(
            (self.x_grid, self.y_grid), seed=4734654, mesh_type="structured"
        )
        self.assertAlmostEqual(field[0, 0, 0], 1.07812013)
        self.assertAlmostEqual(field[0, 1, 0], 1.06180674)

    # TODO put these checks into test_cov_model
    def test_assertions(self):
        # self.cov_model.dim = 0
        # self.assertRaises(ValueError, SRF, self.cov_model, self.mean, self.mode_no)
        # self.cov_model.dim = 4
        # self.assertRaises(ValueError, SRF, self.cov_model, self.mean, self.mode_no)
        self.cov_model.dim = 3
        self.cov_model.anis = (0.25, 0.5)
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        self.assertRaises(ValueError, srf, [self.x_tuple])
        self.assertRaises(ValueError, srf, [self.x_grid, self.y_grid])
        srf = SRF(self.cov_model, mean=self.mean, mode_no=self.mode_no)
        self.assertRaises(ValueError, srf, [self.x_tuple, self.y_tuple])
        self.assertRaises(
            ValueError, srf, [self.x_grid, self.y_grid, self.z_grid]
        )
        self.assertRaises(
            ValueError,
            srf,
            [self.x_tuple, self.y_tuple, self.z_tuple],
            self.seed,
            mesh_type="hyper_mesh",
        )


if __name__ == "__main__":
    unittest.main()



"""Test the PyVista/VTK export methods
"""
import unittest
import numpy as np
import os
import tempfile
import shutil

from gstools import SRF, Gaussian, Exponential
from gstools.random import MasterRNG

HAS_PYVISTA = False
try:
    import pyvista as pv

    HAS_PYVISTA = True
except ImportError:
    pass


class TestKrige(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        # structured field with a size 100x100x100 and a grid-size of 1x1x1
        x = y = z = range(100)
        model = Gaussian(dim=3, var=0.6, len_scale=20)
        self.srf_structured = SRF(model)
        self.srf_structured((x, y, z), mesh_type="structured")
        # unstrucutred field
        seed = MasterRNG(19970221)
        rng = np.random.RandomState(seed())
        x = rng.randint(0, 100, size=10000)
        y = rng.randint(0, 100, size=10000)
        model = Exponential(
            dim=2, var=1, len_scale=[12.0, 3.0], angles=np.pi / 8.0
        )
        self.srf_unstructured = SRF(model, seed=20170519)
        self.srf_unstructured([x, y])

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    @unittest.skipIf(not HAS_PYVISTA, "PyVista is not installed.")
    def test_pyvista(self):
        mesh = self.srf_structured.to_pyvista()
        self.assertIsInstance(mesh, pv.RectilinearGrid)
        mesh = self.srf_unstructured.to_pyvista()
        self.assertIsInstance(mesh, pv.UnstructuredGrid)

    def test_pyevtk_export(self):
        # Structured
        sfilename = os.path.join(self.test_dir, "structured")
        self.srf_structured.vtk_export(sfilename)
        self.assertTrue(os.path.isfile(sfilename + ".vtr"))
        # Unstructured
        ufilename = os.path.join(self.test_dir, "unstructured")
        self.srf_unstructured.vtk_export(ufilename)
        self.assertTrue(os.path.isfile(ufilename + ".vtu"))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
GStools subpackage providing kriging.

.. currentmodule:: gstools.krige

Kriging Classes
^^^^^^^^^^^^^^^

.. autosummary::
   Simple
   Ordinary

----
"""
from __future__ import absolute_import
from gstools.krige.simple import Simple
from gstools.krige.ordinary import Ordinary

__all__ = ["Simple", "Ordinary"]



# -*- coding: utf-8 -*-
"""
GStools subpackage providing upscaling routines for the spatial random field.

.. currentmodule:: gstools.field.upscaling

The following functions are provided

.. autosummary::
   var_coarse_graining
   var_no_scaling
"""
from __future__ import print_function, division, absolute_import

import warnings
import numpy as np

__all__ = ["var_coarse_graining", "var_no_scaling"]


# scaling routines ############################################################


def var_coarse_graining(model, point_volumes=0.0):
    r"""Coarse Graning procedure to upscale the variance for uniform flow.

    Parameters
    ----------
    model : :any:`CovModel`
        Covariance Model used for the field.
    point_volumes : :class:`float` or :class:`numpy.ndarray`
        Volumes of the elements at the given points. Default: ``0``

    Returns
    -------
    scaled_var : :class:`float` or :class:`numpy.ndarray`
        The upscaled variance

    Notes
    -----
    This procedure was presented in [Attinger03]_. It applies the
    upscaling procedure 'Coarse Graining' to the Groundwater flow equation
    under uniform flow on a lognormal distributed conductivity field following
    a gaussian covariance function. A filter over a cube with a given
    edge-length :math:`\lambda` is applied and an upscaled conductivity field
    is obtained.
    The upscaled field is again following a gaussian covariance function with
    scale dependent variance and length-scale:

    .. math::
       \lambda &= V^{\frac{1}{d}} \\
       \sigma^2\left(\lambda\right) &=
       \sigma^2\cdot\left(
       \frac{\ell^2}{\ell^2+\left(\frac{\lambda}{2}\right)^2}
       \right)^{\frac{d}{2}} \\
       \ell\left(\lambda\right) &=
       \left(\ell^2+\left(\frac{\lambda}{2}\right)^2\right)^{\frac{1}{2}}

    Therby :math:`\lambda` will be calculated from the given
    ``point_volumes`` :math:`V` by assuming a cube with the given volume.

    The upscaled length scale will be ignored by this routine.

    References
    ----------
    .. [Attinger03] Attinger, S. 2003,
       ''Generalized coarse graining procedures for flow in porous media'',
       Computational Geosciences, 7(4), 253–273.
    """
    if not np.isclose(model.nugget, 0):
        warnings.warn(
            "var_coarse_graining: non-zero nugget will violate upscaling!"
        )
    # interprete volume as a hypercube and calculate the edge length
    edge = point_volumes ** (1.0 / model.dim)
    var_factor = (
        model.len_scale ** 2 / (model.len_scale ** 2 + edge ** 2 / 4)
    ) ** (model.dim / 2.0)

    return model.sill * var_factor


def var_no_scaling(model, *args, **kwargs):
    r"""Dummy function to bypass scaling.

    Parameters
    ----------
    model : :any:`CovModel`
        Covariance Model used for the field.

    Returns
    -------
    var : :class:`float`
        The model variance.
    """
    return model.sill



# -*- coding: utf-8 -*-
"""
This is the unittest of CovModel class.
"""
from __future__ import division, absolute_import, print_function

import unittest
from gstools import (
    CovModel,
    Gaussian,
    Exponential,
    Rational,
    Stable,
    Matern,
    Linear,
    Circular,
    Spherical,
    TPLGaussian,
    TPLExponential,
    TPLStable,
)


class TestCovModel(unittest.TestCase):
    def setUp(self):
        self.cov_models = [
            Gaussian,
            Exponential,
            Rational,
            Stable,
            Matern,
            Linear,
            Circular,
            Spherical,
            TPLGaussian,
            TPLExponential,
            TPLStable,
        ]
        self.std_cov_models = [
            Gaussian,
            Exponential,
            Rational,
            Stable,
            Matern,
            Linear,
            Circular,
            Spherical,
        ]
        self.dims = range(1, 4)
        self.lens = [[10, 5, 2]]
        self.anis = [[0.5, 0.2]]
        self.nuggets = [0, 1]
        self.vars = [1, 2]
        self.angles = [[1, 2, 3]]

        self.gamma_x = [1.0, 3.0, 5.0, 7.0, 9.0, 11.0]
        self.gamma_y = [0.2, 0.5, 0.6, 0.8, 0.8, 0.9]

    def test_creation(self):
        with self.assertRaises(TypeError):
            CovModel()

        for Model in self.cov_models:
            for dim in self.dims:
                for angles in self.angles:
                    for nugget in self.nuggets:
                        for len_scale, anis in zip(self.lens, self.anis):
                            model = Model(
                                dim=dim, len_scale=len_scale, angles=angles
                            )
                            model1 = Model(
                                dim=dim, len_scale=10, anis=anis, angles=angles
                            )
                            self.assertTrue(model == model1)
                            self.assertAlmostEqual(
                                model.variogram(1),
                                model.var + model.nugget - model.covariance(1),
                            )
                            self.assertAlmostEqual(
                                model.covariance(1),
                                model.var * model.correlation(1),
                            )
                            self.assertAlmostEqual(
                                model.covariance(1),
                                model.var * model.correlation(1),
                            )
                            self.assertAlmostEqual(
                                model.vario_spatial(([1], [2], [3]))[0],
                                model.var
                                + model.nugget
                                - model.cov_spatial(([1], [2], [3]))[0],
                            )
                            self.assertAlmostEqual(
                                model.cov_nugget(0), model.sill
                            )
                            self.assertAlmostEqual(model.vario_nugget(0), 0.0)
                            self.assertAlmostEqual(
                                model.cov_nugget(1), model.covariance(1)
                            )
                            self.assertAlmostEqual(model.vario_nugget(0), 0.0)
                            self.assertAlmostEqual(
                                model.vario_nugget(1), model.variogram(1)
                            )
                            # check if callable
                            model.vario_spatial((1, 2, 3))
                            model.spectral_density([0, 1])
                            model.spectrum([0, 1])
                            model.spectral_rad_pdf([0, 1])
                            model.ln_spectral_rad_pdf([0, 1])
                            model.integral_scale_vec
                            model.percentile_scale(0.9)
                            if model.has_cdf:
                                model.spectral_rad_cdf([0, 1])
                            if model.has_ppf:
                                model.spectral_rad_ppf([0.0, 0.99])
                            model.pykrige_kwargs

    def test_fitting(self):
        for Model in self.std_cov_models:
            for dim in self.dims:
                model = Model(dim=dim)
                model.fit_variogram(self.gamma_x, self.gamma_y, nugget=False)
                self.assertAlmostEqual(model.nugget, 0.0)


if __name__ == "__main__":
    unittest.main()
